from typing import List

from py import test
from spacy.tokens import Doc, Span


class CorefResolver(object):
    """ a class that implements the logic from
        https://towardsdatascience.com/how-to-make-an-effective-coreference-resolution-model-55875d2b5f19"""
    def __init__(self) -> None:
        pass
       
    @staticmethod 
    def core_logic_part(document: Doc, coref: List[int], resolved: List[str], mention_span: Span):
        final_token = document[coref[1]]
        final_token_tag = str(final_token.tag_).lower()
        test_token_test = False
        for option in ["PRP$", "POS", 'BEZ']:
            if option.lower() in final_token_tag:
                test_token_test = True
                break
        if test_token_test:
            resolved[coref[0]] = mention_span.text + "'s" + final_token.whitespace_
        else:
            resolved[coref[0]] = mention_span.text + final_token.whitespace_
        for i in range(coref[0] + 1, coref[1] + 1):
            resolved[i] = ""
        return resolved

    @staticmethod
    def get_span_noun_indices(doc: Doc, cluster: List[List[int]]) -> List[int]:
        spans = [doc[span[0]:span[1]+1] for span in cluster]
        spans_pos = [[token.pos_ for token in span] for span in spans]
        span_noun_indices = [i for i, span_pos in enumerate(spans_pos)
            if any(pos in span_pos for pos in ['NOUN', 'PROPN'])]
        return span_noun_indices

    @staticmethod
    def get_cluster_head(doc: Doc, cluster: List[List[int]], noun_indices: List[int]):
        head_idx = noun_indices[0]
        head_start, head_end = cluster[head_idx]
        head_span = doc[head_start:head_end+1]
        return head_span, [head_start, head_end]

    @staticmethod
    def is_containing_other_spans(span: List[int], all_spans: List[List[int]]):
        return any([s[0] >= span[0] and s[1] <= span[1] and s != span for s in all_spans])

    def replace_corefs(self, document, clusters):
        resolved = list(tok.text_with_ws for tok in document)
        all_spans = [span for cluster in clusters for span in cluster]  # flattened list of all spans

        for cluster in clusters:
            noun_indices = self.get_span_noun_indices(document, cluster)

            if noun_indices:
                mention_span, mention = self.get_cluster_head(document, cluster, noun_indices)

                for coref in cluster:
                    if coref != mention and not self.is_containing_other_spans(coref, all_spans):
                        self.core_logic_part(document, coref, resolved, mention_span)
        return "".join(resolved)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['crosslingual_coreference', 'crosslingual_coreference.examples']

package_data = \
{'': ['*']}

install_requires = \
['allennlp-models>=2.7,<3.0',
 'allennlp>=2.7,<3.0',
 'click<8.1.0',
 'spacy>=3.1.0,<4.0']

setup_kwargs = {
    'name': 'crosslingual-coreference',
    'version': '0.1.5',
    'description': 'A multi-lingual approach to AllenNLP CoReference Resolution, along with a wrapper for spaCy.',
    'long_description': '# Crosslingual Coreference\nCoreference is amazing but the data required for training a model is very scarce. In our case, the available training for non-English languages also proved to be poorly annotated. Crosslingual Coreference, therefore, uses the assumption a trained model with English data and cross-lingual embeddings should work for languages with similar sentence structures.\n\n[![Current Release Version](https://img.shields.io/github/release/pandora-intelligence/crosslingual-coreference.svg?style=flat-square&logo=github)](https://github.com/pandora-intelligence/crosslingual-coreference/releases)\n[![pypi Version](https://img.shields.io/pypi/v/crosslingual-coreference.svg?style=flat-square&logo=pypi&logoColor=white)](https://pypi.org/project/crosslingual-coreference/)\n[![PyPi downloads](https://static.pepy.tech/personalized-badge/crosslingual-coreference?period=total&units=international_system&left_color=grey&right_color=orange&left_text=pip%20downloads)](https://pypi.org/project/crosslingual-coreference/)\n# Install\n\n```\npip install crosslingual-coreference\n```\n# Quickstart\n```python\nfrom crosslingual_coreference import Predictor\n\ntext = "Do not forget about Momofuku Ando! He created instant noodles in Osaka. At that location, Nissin was founded. Many students survived by eating these noodles, but they don\'t even know him."\n\npredictor = Predictor(language="en_core_web_sm", device=-1, model_name="info_xlm")\n\nprint(predictor.predict(text)["resolved_text"])\n# Output\n# \n# Do not forget about Momofuku Ando! \n# Momofuku Ando created instant noodles in Osaka. \n# At Osaka, Nissin was founded. \n# Many students survived by eating instant noodles, \n# but Many students don\'t even know Momofuku Ando.\n```\n![](https://raw.githubusercontent.com/Pandora-Intelligence/crosslingual-coreference/master/img/example_en.png)\n## Use spaCy pipeline\n```python\nimport crosslingual_coreference\nimport spacy\n\ntext = "Do not forget about Momofuku Ando! He created instant noodles in Osaka. At that location, Nissin was founded. Many students survived by eating these noodles, but they don\'t even know him."\n\nnlp = spacy.load(\'en_core_web_sm\')\nnlp.add_pipe(\'xx_coref\')\n\ndoc = nlp(text)\nprint(doc._.coref_clusters)\n# Output\n# \n# [[[4, 5], [7, 7], [27, 27], [36, 36]], \n# [[12, 12], [15, 16]], \n# [[9, 10], [27, 28]], \n# [[22, 23], [31, 31]]]\nprint(doc._.resolved_text)\n# Output\n# \n# Do not forget about Momofuku Ando! \n# Momofuku Ando created instant noodles in Osaka. \n# At Osaka, Nissin was founded. \n# Many students survived by eating instant noodles, \n# but Many students don\'t even know Momofuku Ando.\n```\n## Available models\nAs of now, there are two models available "info_xlm", "xlm_roberta", which scored 77 and 74 on OntoNotes Release 5.0 English data, respectively.\n## More Examples\n![](https://raw.githubusercontent.com/Pandora-Intelligence/crosslingual-coreference/master/img/example_total.png)\n\n',
    'author': 'David Berenstein',
    'author_email': 'david.berenstein@pandoraintelligence.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pandora-intelligence/crosslingual-coreference',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0',
}


setup(**setup_kwargs)



import pathlib
from typing import List

import requests
import tqdm  # progress bar
from allennlp.predictors.predictor import Predictor

from .CorefResolver import CorefResolver as Resolver

MODELS = {
    'xlm_roberta': {
        'url': 'https://storage.googleapis.com/pandora-intelligence/models/crosslingual-coreference/xlm-roberta-base/model.tar.gz',
        'f1_score_ontonotes': 74,
        'file_extension': '.tar.gz'
    },
    'info_xlm': {
        'url': 'https://storage.googleapis.com/pandora-intelligence/models/crosslingual-coreference/infoxlm-base/model.tar.gz',
        'f1_score_ontonotes': 77,
        'file_extension': '.tar.gz'
    }
}


class CrossLingualPredictor(object):
    def __init__(self, language: str, device: int = -1, model_name: str = 'info_xlm') -> None:
        self.language = language
        self.filename = None
        self.device = device
        self.model_url = MODELS[model_name]['url']
        self.resolver = Resolver()
        self.download_model()
        self.set_coref_model()
    
    def download_model(self):
        """
        Download file with progressbar if file is present, otherwise pass
        """
        self.filename = self.model_url.replace('https://storage.googleapis.com/pandora-intelligence/', '')
        path = pathlib.Path(self.filename)
        if path.is_file():
            pass
        else:
            path.parent.absolute().mkdir(parents=True, exist_ok=True)
            r = requests.get(self.model_url, stream=True)
            file_size = int(r.headers['Content-Length'])
            
            chunk_size = 1024
            num_bars = int(file_size / chunk_size)

            with open(self.filename, 'wb') as fp:
                for chunk in tqdm.tqdm(
                                        r.iter_content(chunk_size=chunk_size),
                                        total = num_bars,
                                        unit = 'KB',
                                        desc = self.filename,
                                        leave = True 
                                    ):
                    fp.write(chunk)
    
    def set_coref_model(self):
        """Initialize AllenNLP coreference model """
        self.predictor = Predictor.from_path(self.filename, language=self.language, cuda_device=self.device)

    def predict(self, text: str, advanced_resolve: bool = True) -> dict:
        """predict and rsolve

        Args:
            text (str): an input text
            advanced_resolve (bool, optional): use more advanced resoled from 
            https://towardsdatascience.com/how-to-make-an-effective-coreference-resolution-model-55875d2b5f19. Defaults to True.

        Returns:
            dict: a prediciton
        """
        prediction = self.predictor.predict_json({"document": text})
        
        clusters = prediction.get("clusters")

        # Passing a document with no coreferences returns its original form
        if not clusters:
            prediction['resolved_text'] = text
            return prediction
        
        doc = self.predictor._spacy(text)
        
        if advanced_resolve:
            prediction['resolved_text'] = self.resolver.replace_corefs(doc, clusters)
        else:
            prediction['resolved_text'] = self.predictor.replace_corefs(doc, clusters)
            
        return prediction
        
    def pipe(self, texts: List[str], advanced_resolve: bool = True) -> List[dict]:
        return [self.predict(text, advanced_resolve) for text in texts]
